@@ -1,8 +1,11 @@
+import contextlib
+import enum
 import functools
 import logging
 import typing
 from collections import deque
-from typing import Any, Callable, ParamSpec, TypeVar
+from collections.abc import Callable, Sequence
+from typing import Any, ParamSpec, TypeVar
 
 import anyio
 import attrs
@@ -15,8 +18,14 @@ from beartype.roar import BeartypeCallHintViolation
 from cattrs.preconf.json import make_converter
 from gql.client import AsyncClientSession, SyncClientSession
 from gql.dsl import DSLField, DSLQuery, DSLSchema, DSLSelectable, DSLType, dsl_gql
+from gql.transport.exceptions import TransportQueryError
 
-from dagger.exceptions import ExecuteTimeoutError, InvalidQueryError
+from dagger.exceptions import (
+    ExecuteTimeoutError,
+    InvalidQueryError,
+    QueryError,
+    QueryErrorLocation,
+)
 
 logger = logging.getLogger(__name__)
 
@@ -42,12 +51,21 @@ class IDType(typing.Protocol):
         ...
 
 
+@attrs.define
+class _QueryError:
+    message: str
+    locations: list[QueryErrorLocation]
+    path: list[str]
+
+
 @attrs.define
 class Context:
     session: AsyncClientSession | SyncClientSession
     schema: DSLSchema
     selections: deque[Field] = attrs.field(factory=deque)
-    converter: cattrs.Converter = attrs.field(factory=make_converter)
+    converter: cattrs.Converter = attrs.field(
+        factory=functools.partial(make_converter, detailed_validation=False)
+    )
 
     def select(
         self,
@@ -83,52 +101,87 @@ class Context:
         assert isinstance(self.session, AsyncClientSession)
         await self.resolve_ids()
         query = self.query()
-        try:
-            result = await self.session.execute(query, get_execution_result=True)
-        except httpx.TimeoutException as e:
-            msg = (
-                "Request timed out. Try setting a higher value in 'execute_timeout' "
-                "config for this `dagger.Connection()`."
-            )
-            raise ExecuteTimeoutError(msg) from e
-        return self._get_value(result.data, return_type)
+        with self._handle_execute(query):
+            result = await self.session.execute(query)
+        return self.get_value(result, return_type)
 
     def execute_sync(self, return_type: type[T]) -> T:
         assert isinstance(self.session, SyncClientSession)
         self.resolve_ids_sync()
         query = self.query()
+        with self._handle_execute(query):
+            result = self.session.execute(query)
+        return self.get_value(result, return_type)
+
+    async def resolve_ids(self) -> None:
+        """Replace Type object instances with their ID implicitly."""
+
+        # mutating to avoid re-fetching on forked pipeline
+        async def _resolve_id(pos: int, k: str, v: IDType):
+            sel = self.selections[pos]
+            sel.args[k] = await v.id()
+
+        async def _resolve_seq_id(pos: int, idx: int, k: str, v: IDType):
+            sel = self.selections[pos]
+            sel.args[k][idx] = await v.id()
+
+        # resolve all ids concurrently
+        async with anyio.create_task_group() as tg:
+            for i, sel in enumerate(self.selections):
+                for k, v in sel.args.items():
+                    # check if it's a sequence of Type objects
+                    if TypeSequence.is_bearable(v):
+                        # make sure it's a list, to mutate by index
+                        sel.args[k] = v = list(v)
+                        for seq_i, seq_v in enumerate(v):
+                            if isinstance(seq_v, IDType):
+                                tg.start_soon(_resolve_seq_id, i, seq_i, k, seq_v)
+                    elif isinstance(v, (Type, IDType)):
+                        tg.start_soon(_resolve_id, i, k, v)
+
+    def resolve_ids_sync(self) -> None:
+        """Replace Type object instances with their ID implicitly."""
+        for sel in self.selections:
+            for k, v in sel.args.items():
+                # check if it's a sequence of Type objects
+                if TypeSequence.is_bearable(v):
+                    # make sure it's a list, to mutate by index
+                    sel.args[k] = v = list(v)
+                    for seq_i, seq_v in enumerate(v):
+                        if isinstance(seq_v, IDType):
+                            sel.args[k][seq_i] = seq_v.id()
+                elif isinstance(v, (Type, IDType)):
+                    sel.args[k] = v.id()
+
+    @contextlib.contextmanager
+    def _handle_execute(self, query: graphql.DocumentNode):
         try:
-            result = self.session.execute(query, get_execution_result=True)
+            yield
         except httpx.TimeoutException as e:
             msg = (
                 "Request timed out. Try setting a higher value in 'execute_timeout' "
                 "config for this `dagger.Connection()`."
             )
             raise ExecuteTimeoutError(msg) from e
-        return self._get_value(result.data, return_type)
+        except TransportQueryError as e:
+            if error := self._parse_query_error(e):
+                raise QueryError(
+                    error.message.strip(),
+                    query,
+                    error.path,
+                    error.locations,
+                ) from e
+            raise
 
-    async def resolve_ids(self) -> None:
-        # mutating to avoid re-fetching on forked pipeline
-        async def _resolve_id(pos: int, k: str, v: IDType):
-            sel = self.selections[pos]
-            sel.args[k] = await v.id()
+    def _parse_query_error(self, exc: TransportQueryError) -> _QueryError | None:
+        try:
+            return self.converter.structure(exc.errors, list[_QueryError])[0]
+        except (TypeError, KeyError, ValueError, IndexError):
+            return None
 
-        # resolve all ids concurrently
-        async with anyio.create_task_group() as tg:
-            for i, sel in enumerate(self.selections):
-                for k, v in sel.args.items():
-                    if isinstance(v, (Type, IDType)):
-                        tg.start_soon(_resolve_id, i, k, v)
-
-    def resolve_ids_sync(self) -> None:
-        for sel in self.selections:
-            for k, v in sel.args.items():
-                if isinstance(v, (Type, IDType)):
-                    sel.args[k] = v.id()
-
-    def _get_value(self, value: dict[str, Any] | None, return_type: type[T]) -> T:
+    def get_value(self, value: dict[str, Any] | None, return_type: type[T]) -> T:
         if value is not None:
-            value = self._structure_response(value, return_type)
+            value = self.structure_response(value, return_type)
         if value is None and not TypeHint(return_type).is_bearable(None):
             msg = (
                 "Required field got a null response. Check if parent fields are valid."
@@ -136,7 +189,7 @@ class Context:
             raise InvalidQueryError(msg)
         return value
 
-    def _structure_response(self, response: dict[str, Any], return_type: type[T]) -> T:
+    def structure_response(self, response: dict[str, Any], return_type: type[T]) -> T:
         for f in self.selections:
             response = response[f.name]
             if response is None:
@@ -154,6 +207,13 @@ class Scalar(str):
     """Custom scalar."""
 
 
+class Enum(str, enum.Enum):
+    """Custom enumeration."""
+
+    def __str__(self) -> str:
+        return str(self.value)
+
+
 class Object:
     """Base for object types."""
 
@@ -194,6 +254,9 @@ class Root(Type):
         return "Query"
 
 
+TypeSequence = TypeHint(Sequence[Type])
+
+
 def typecheck(func: Callable[P, T]) -> Callable[P, T]:
     """
     Runtime type checking.
